@@ -211,11 +211,17 @@ def spa_worker(config, q_spa, pauseonload):
             done.append(action)
             if action == 'objstore':
                 altrepo = grokmirror.get_altrepo(fullpath)
-                grokmirror.fetch_objstore_repo(altrepo, fullpath)
+                # Should we use plumbing for this?
+                use_plumbing = config['core'].getboolean('objstore_uses_plumbing', False)
+                grokmirror.fetch_objstore_repo(altrepo, fullpath, use_plumbing=use_plumbing)
 
             elif action == 'repack':
                 logger.debug('quick-repacking %s', fullpath)
                 args = ['repack', '-Adlq']
+                if 'fsck' in config:
+                    extraflags = config['fsck'].get('extra_repack_flags', '').split()
+                    if len(extraflags):
+                        args += extraflags
                 ecode, out, err = grokmirror.run_git_command(fullpath, args)
                 if ecode > 0:
                     logger.debug('Could not repack %s', fullpath)
@@ -236,12 +242,57 @@ def spa_worker(config, q_spa, pauseonload):
         logger.info('      spa: %s (done: %s)', gitdir, ', '.join(done))
 
 
+def objstore_repo_preload(config, obstrepo):
+    purl = config['remote'].get('preload_bundle_url')
+    if not purl:
+        return
+    bname = os.path.basename(obstrepo)[:-4]
+    obstdir = os.path.realpath(config['core'].get('objstore'))
+    burl = '%s/%s.bundle' % (purl.rstrip('/'), bname)
+    bfile = os.path.join(obstdir, '%s.bundle' % bname)
+    try:
+        sess = grokmirror.get_requests_session()
+        resp = sess.get(burl, stream=True)
+        resp.raise_for_status()
+        logger.info(' objstore: downloading %s.bundle', bname)
+        with open(bfile, 'wb') as fh:
+            for chunk in resp.iter_content(chunk_size=8192):
+                fh.write(chunk)
+        resp.close()
+    except: # noqa
+        # Make sure we don't leave .bundle files lying around
+        # Should we add logic to resume downloads here in the future?
+        if os.path.exists(bfile):
+            os.unlink(bfile)
+        return
+
+    # Now we clone from it into the objstore repo
+    ecode, out, err = grokmirror.run_git_command(obstrepo, ['remote', 'add', '--mirror=fetch', '_preload', bfile])
+    if ecode == 0:
+        logger.info(' objstore: preloading %s.bundle', bname)
+        args = ['remote', 'update', '_preload']
+        ecode, out, err = grokmirror.run_git_command(obstrepo, args)
+        if ecode > 0:
+            logger.info(' objstore: failed to preload from %s.bundle', bname)
+        else:
+            # now pack refs and generate a commit graph
+            grokmirror.run_git_command(obstrepo, ['pack-refs', '--all'])
+            if grokmirror.git_newer_than('2.18.0'):
+                grokmirror.run_git_command(obstrepo, ['commit-graph', 'write'])
+            logger.info(' objstore: successful preload from %s.bundle', bname)
+    # Regardless of what happened, we remove _preload and the bundle, then move on
+    grokmirror.run_git_command(obstrepo, ['remote', 'rm', '_preload'])
+    os.unlink(bfile)
+
+
 def pull_worker(config, q_pull, q_spa, q_done):
     toplevel = os.path.realpath(config['core'].get('toplevel'))
     obstdir = os.path.realpath(config['core'].get('objstore'))
     maxretries = config['pull'].getint('retries', 3)
     site = config['remote'].get('site')
     remotename = config['pull'].get('remotename', '_grokmirror')
+    # Should we use plumbing for objstore operations?
+    objstore_uses_plumbing = config['core'].getboolean('objstore_uses_plumbing', False)
 
     while True:
         try:
@@ -263,6 +314,11 @@ def pull_worker(config, q_pull, q_spa, q_done):
             q_pull.put((gitdir, repoinfo, action, q_action))
             continue
 
+        altrepo = grokmirror.get_altrepo(fullpath)
+        obstrepo = None
+        if altrepo and grokmirror.is_obstrepo(altrepo, obstdir):
+            obstrepo = altrepo
+
         if action == 'purge':
             # Is it a symlink?
             if os.path.islink(fullpath):
@@ -276,19 +332,19 @@ def pull_worker(config, q_pull, q_spa, q_done):
                     logger.info('    purge: %s', gitdir)
                     shutil.rmtree(fullpath)
 
+        if action == 'fix_params':
+            logger.info(' reconfig: %s', gitdir)
+            set_repo_params(fullpath, repoinfo)
+
         if action == 'fix_remotes':
             logger.info(' reorigin: %s', gitdir)
             success = fix_remotes(toplevel, gitdir, site, config)
             if success:
-                action = 'fix_params'
+                set_repo_params(fullpath, repoinfo)
+                action = 'pull'
             else:
                 success = False
 
-        if action == 'fix_params':
-            logger.info(' reconfig: %s', gitdir)
-            set_repo_params(fullpath, repoinfo)
-            action = 'pull'
-
         if action == 'reclone':
             logger.info('  reclone: %s', gitdir)
             try:
@@ -308,6 +364,11 @@ def pull_worker(config, q_pull, q_spa, q_done):
         if action in ('pull', 'objstore_migrate'):
             r_fp = repoinfo.get('fingerprint')
             my_fp = grokmirror.get_repo_fingerprint(toplevel, gitdir, force=True)
+            if obstrepo:
+                o_obj_info = grokmirror.get_repo_obj_info(obstrepo)
+                if o_obj_info.get('count') == '0' and o_obj_info.get('in-pack') == '0' and not my_fp:
+                    # Try to preload the objstore repo directly
+                    objstore_repo_preload(config, obstrepo)
 
             if r_fp != my_fp:
                 # Make sure we have the remote set up
@@ -317,7 +378,7 @@ def pull_worker(config, q_pull, q_spa, q_done):
                 logger.info('    fetch: %s', gitdir)
                 retries = 1
                 while True:
-                    success = pull_repo(toplevel, gitdir, remotename)
+                    success = pull_repo(fullpath, remotename)
                     if success:
                         break
                     retries += 1
@@ -338,12 +399,13 @@ def pull_worker(config, q_pull, q_spa, q_done):
                             if o_obj_info.get('count') == '0' and o_obj_info.get('in-pack') == '0':
                                 # We fetch right now, as other repos may be waiting on these objects
                                 logger.info(' objstore: %s', gitdir)
-                                grokmirror.fetch_objstore_repo(altrepo, fullpath)
-                                spa_actions.append('repack')
+                                grokmirror.fetch_objstore_repo(altrepo, fullpath, use_plumbing=objstore_uses_plumbing)
+                                if not objstore_uses_plumbing:
+                                    spa_actions.append('repack')
                             else:
                                 # We lazy-fetch in the spa
                                 spa_actions.append('objstore')
-                                if my_fp is None:
+                                if my_fp is None and not objstore_uses_plumbing:
                                     # Initial clone, trigger a repack after objstore
                                     spa_actions.append('repack')
 
@@ -534,8 +596,7 @@ def run_post_update_hook(toplevel, gitdir, hookscripts):
             logger.info('Hook Stdout (%s): %s', gitdir, output)
 
 
-def pull_repo(toplevel, gitdir, remotename):
-    fullpath = os.path.join(toplevel, gitdir.lstrip('/'))
+def pull_repo(fullpath, remotename):
     args = ['remote', 'update', remotename, '--prune']
 
     retcode, output, error = grokmirror.run_git_command(fullpath, args)
@@ -562,9 +623,9 @@ def pull_repo(toplevel, gitdir, remotename):
             else:
                 debug.append(line)
         if debug:
-            logger.debug('Stderr (%s): %s', gitdir, '\n'.join(debug))
+            logger.debug('Stderr (%s): %s', fullpath, '\n'.join(debug))
         if warn:
-            logger.warning('Stderr (%s): %s', gitdir, '\n'.join(warn))
+            logger.warning('Stderr (%s): %s', fullpath, '\n'.join(warn))
 
     return success
 
@@ -824,6 +885,14 @@ def fill_todo_from_manifest(config, q_mani, nomtime=False, forcepurge=False):
             if r_desc != l_desc or r_owner != l_owner or r_head != l_head:
                 q_mani.put((gitdir, repoinfo, 'fix_params'))
 
+            if symlinks and isinstance(symlinks, list):
+                # Are all symlinks in place?
+                for symlink in symlinks:
+                    linkpath = os.path.join(toplevel, symlink.lstrip('/'))
+                    if not os.path.islink(linkpath) or os.path.realpath(linkpath) != fullpath:
+                        q_mani.put((gitdir, repoinfo, 'fix_params'))
+                        break
+
             my_fingerprint = grokmirror.get_repo_fingerprint(toplevel, gitdir)
             if my_fingerprint != l_manifest[gitdir].get('fingerprint'):
                 logger.debug('Fingerprint discrepancy, forcing a fetch')
@@ -912,7 +981,15 @@ def fill_todo_from_manifest(config, q_mani, nomtime=False, forcepurge=False):
                     if fnmatch.fnmatch(gitdir, entry):
                         exclude = True
                         break
+                # Refuse to purge ffonly repos
+                for globpatt in set([x.strip() for x in config['pull'].get('ffonly', '').split('\n')]):
+                    if fnmatch.fnmatch(gitdir, globpatt):
+                        # Woah, these are not supposed to be deleted, ever
+                        logger.critical('Refusing to purge ffonly repo %s', gitdir)
+                        exclude = True
+                        break
                 if not exclude:
+                    logger.debug('Adding %s to to_purge', gitdir)
                     to_purge.add(gitdir)
 
         if len(to_purge):
@@ -933,6 +1010,7 @@ def fill_todo_from_manifest(config, q_mani, nomtime=False, forcepurge=False):
                 logger.critical('Set purgeprotect to a higher percentage, or override with --force-purge.')
             else:
                 for gitdir in to_purge:
+                    logger.debug('Queued %s for purging', gitdir)
                     q_mani.put((gitdir, None, 'purge'))
         else:
             logger.debug('No repositories need purging')
@@ -1053,7 +1131,7 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
     # Run in the main thread if we have runonce
     if runonce:
         fill_todo_from_manifest(config, q_mani, nomtime=nomtime, forcepurge=forcepurge)
-        if q_mani.empty():
+        if not q_mani.qsize():
             return 0
     else:
         # force nomtime to True the first time
@@ -1061,9 +1139,11 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
     lastrun = 0
 
     pull_threads = config['pull'].getint('pull_threads', 0)
-    if pull_threads < 1:
+    if pull_threads < 1 and mp.cpu_count() > 1:
         # take half of available CPUs by default
         pull_threads = int(mp.cpu_count() / 2)
+    else:
+        pull_threads = 1
 
     busy = set()
     done = list()
@@ -1180,8 +1260,12 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
                                             dw.join()
                                         return 0
                                     time.sleep(1)
-                    # Don't run a hot loop waiting on results
-                    time.sleep(1)
+                    if len(pws):
+                        # Don't run a hot loop waiting on results
+                        time.sleep(5)
+                    else:
+                        # Shorter sleep if everything is idle
+                        time.sleep(1)
                 continue
 
             if repoinfo is None:
@@ -1195,8 +1279,9 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
                 if loopmark is None:
                     loopmark = gitdir
                 elif loopmark == gitdir:
-                    # sleep for a bit if we recognize that we've looped around all waiting repos
-                    time.sleep(0.5)
+                    # We've looped around all waiting repos, so back off and don't run
+                    # a hot waiting loop.
+                    time.sleep(5)
                 continue
 
             if gitdir == loopmark:
